
import unittest
import sys
from ..clipboard import Clipboard
from .. import platform_backend
from PIL import Image as PilImage
from PIL import ImageChops as PilImageChops
import numpy

def generate_random_image(image_format='RGB'):
    imarray = numpy.random.rand(100, 100, 3) * 255
    # TODO: for now, clipboard does not support RGBA. This will need some
    # work
    test_image = PilImage.fromarray(imarray.astype('uint8')).convert(image_format)
    return test_image

def eval_images(image1, image2):
    """ Function to verify that two images are the same
    """
    return PilImageChops.difference(image1, image2).getbbox() is None

@unittest.skipUnless(platform_backend == 'gtk', 'Not using GTK backend')
class GtkTestCase(unittest.TestCase):

    def setUp(self):
        # Create clipboard. Make sure that it is valid
        self.clipboard = Clipboard()
        self.assertTrue(self.clipboard is not None)

    def test_text(self):
        # Test putting text onto the clipboard
        msg = 'Hello World'
        self.clipboard.set_text(msg)

        # Test getting above text
        text = self.clipboard.get_text()
        self.assertTrue(text is not None)
        self.assertTrue(text == msg)

    def test_image(self):
        # Open the test image
        test_image = generate_random_image()

        # Put image on the clipboard
        self.clipboard.set_image(test_image)

        # Get image from clipboard, verify equality
        new_image = self.clipboard.get_image()
        self.assertTrue(
            eval_images(test_image, new_image)
        )

@unittest.skipUnless(platform_backend == 'qt', 'Not using Qt backend')
class QtTestCase(unittest.TestCase):

    def setUp(self):
        # Create clipboard. Make sure that it is valid
        self.clipboard = Clipboard()
        self.assertTrue(self.clipboard is not None)

    def test_text(self):
        # Test putting text onto the clipboard
        msg = 'Hello World'
        self.clipboard.set_text(msg)

        # Test getting above text
        text = self.clipboard.get_text()
        self.assertTrue(text is not None)
        self.assertTrue(text == msg)

    def test_image(self):
        # Open the test image
        test_image = generate_random_image()

        # Put image on the clipboard
        self.clipboard.set_image(test_image)

        # Get image from clipboard, verify equality
        new_image = self.clipboard.get_image()
        self.assertTrue(
            eval_images(test_image, new_image)
        )

@unittest.skipUnless(platform_backend == 'win32', 'Not using windows backend')
class WinTestCase(unittest.TestCase):

    def setUp(self):
        # Create clipboard. Make sure that it is valid
        self.clipboard = Clipboard()
        self.assertTrue(self.clipboard is not None)

    def test_text(self):
        # Test putting text onto the clipboard
        msg = 'Hello World'
        self.clipboard.set_text(msg)

        # Test getting above text
        text = self.clipboard.get_text()
        self.assertTrue(text is not None)
        self.assertTrue(text == msg)

    def test_image(self):
        # Open the test image
        test_image = generate_random_image()

        # Put image on the clipboard
        self.clipboard.set_image(test_image)

        # Get image from clipboard, verify equality
        new_image = self.clipboard.get_image()
        self.assertTrue(
            eval_images(test_image, new_image)
        )




# crossclip -- cross platform clipboard API
# Copyright (C) 2019  Charlie Sale

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# backends.py -- backend classes

import sys
import os
from abc import ABC, abstractmethod, abstractstaticmethod, abstractproperty

class AbstractBackend(ABC):
    """ Interface for all clipboard backends

    This class is an interface for all clipboard backends to
    be registered into.
    """
    @abstractmethod
    def get_text(self):
        """ Synchronously gets text from clipboard
        :returns: Text from clipboard
        :rtype: str
        """
        pass

    @abstractmethod
    def get_image(self, form):
        """ Synchronously gets image from clipboard
        :returns: Image from clipboard
        :rtype:
        """
        pass

    @abstractmethod
    def set_text(self, text):
        """ Sets text to the clipboard
        :param text: Text to add
        :type text: str
        """
        pass

    @abstractmethod
    def set_image(self, img):
        """ Sets image to clipboard
        :param img: Image to set to clipboard
        :type img:
        """
        pass

class AbstractImageConverter(ABC):
    """ Converts an image between a Pillow Image and a native clipboard image

    This interface is needed to convert native images to pillow images and vice
    versa.
    """
    @abstractproperty
    def image_type(self):
        """ The type of native image
        """
        pass

    @abstractproperty
    def image_str(self):
        """ String representing image
        """
        pass

    @abstractmethod
    def to_pillow(self, native_image):
        """ Converts a native image to Pillow iamge
        """
        pass

    @abstractmethod
    def from_pillow(self, pil):
        """ Converts a pillow image to native type
        """
        pass




# crossclip -- cross platform clipboard API
# Copyright (C) 2019  Charlie Sale

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# clipboard.py -- frontend clipboard class

import sys
from . import platform_backend, backends
from .absbackend import AbstractBackend
import PIL

class Clipboard:
    """ Frontend to various clipboard backends

    This class is a frontend for the various backends. It automatically
    determines which backend should be used based on whatever the platform
    is.
    """
    backend_type = None
    """ Type object of backend.
    """
    backend = None
    """ Actual backend instance
    """
    image_converter = None
    """ Image converter instance
    """

    def __init__(self, clip_backend_type=backends[platform_backend]):
        """
        Creates a new clipboard that interfaces one of the platform-specific
        backends. The backend is implicitly deduced, but a specific backend
        can be choosen instead.

        :param clip_backend: Which backend to use. Defaults to implicitly-selected backend
        :type clip_backend: instance of `AbstractBackend`
        :raises RuntimeError: If clip_backend is invalid
        """
        # Choose the backend to use
        #if clip_backend not in backends:
        #    raise RuntimeError('Invalid backend selected')

        # Verify validity of backend type
        if not issubclass(clip_backend_type, AbstractBackend):
            raise RuntimeError("Clipboard backend is of invalid type")

        self.backend = clip_backend_type()
        self.backend_type = clip_backend_type

        # Based off of backend, get the native image type (e.g QImage)
        self.image_converter = self.backend.image_converter

    def get_text(self):
        """
        Gets text from the clipboard.

        :returns: Text from clipboard or None if no text is available
        :rtype: str
        """
        return self.backend.get_text()

    def get_image(self, form='pil', converter=None):
        """
        Gets an image from the clipboard.

        :param native: If true, then the returned image will be of type `self.image_converter.image_type`.
                        If False, then object will be of type `PIL.Image`.
        :type native: boolean
        :returns: Initialized image object or None if no image is available
        :rtype: `PIL.Image` or `self.image_converter.image_type`
        """
        return self.backend.get_image(form, converter)

    def set_text(self, text: str):
        """
        Places text on the clipboard.

        :param text: Text to add
        :type text: str
        """
        self.backend.set_text(text)

    def set_image(self, image):
        """
        Sets an image on the clipboard. Image can either be of type `PIL.Image` or
        `self.image_converter.image_type`.

        :param image: image to be placed.
        :type image: instance of `PIL.Image` or `self.image_converter.image_type`
        :raises RuntimeError: If image is neither of type `PIL.Image` nor `self.image_converter.image_type`
        """
        self.backend.set_image(image)





import sys
import os

# Do some cross-platform importing. This module does not support
# cygwin. 
# On linux, the system first looks for Gtk via PyGObject.
# If that is not found, then it tries to get Qt. If neither are
# found, then an error is thrown. Ill try to support more formats
# in the future, but as of now, I am operating under the premise
# that everyone that would be using this on linux is using either
# a GTK or a Qt based desktop.
# On OSX, I use PyObjC to access the system clipboard. I will consider
# adding fallback support to use pb(copy|paste) as a backup.
# On windows, use the win32clipboard module. If that's not found, then
# no dice.

platform_backend = None

backends = {
    'gtk': None,
    'qt': None,
    'apple': None,
    'win': None
}

if sys.platform == 'linux':
    # Get current desktop
    current_desktop = os.environ.get('XDG_CURRENT_DESKTOP')
    if current_desktop in ['MATE', 'GNOME', 'X-Cinnamon', 'LXDE', 'XFCE', 'Unity']:
        # USE GTK AS BACKEND
        platform_backend = 'gtk'
        from .gtkbackend import GtkBackend
        backends['gtk'] = GtkBackend
    elif current_desktop in ['LXQt', 'KDE', ]:
        # USE QT AS BACKEND
        platform_backend = 'qt'
        from .qtbackend import QtBackend
        backends['qt'] = QtBackend
    else:
        # This block checks to see if using the Xfce4 desktop
        try:
            pipe = os.popen('xprop -root _DT_SAVE_MODE')
            if ' = "xfce4"' in pipe.read():
                platform_backend = 'gtk'
                from .gtkbackend import GtkBackend
                backends['gtk'] = GtkBackend
            pipe.close()
        except (OSError, RuntimeError):
            raise RuntimeError('Not using a GTK or Qt-based Desktop')

elif sys.platform == 'darwin':
    try:
        # Try getting PyObjC
        from AppKit import NSPasteboard, NSStringPboardType
        from PIL import ImageGrab as PilImageGrab

        platform_backend = 'apple'
    except ModuleNotFoundError:
        raise RuntimeError("You need PyObjC if you are running on mac")

elif sys.platform == 'win32':
    try:
        platform_backend = 'win'
        from .winbackend import WindowsBackend
        backends['win'] = WindowsBackend
    except ModuleNotFoundError:
        raise RuntimeError("You need pywin32 to run on windows")
else:
    raise RuntimeError('Your platform is not supported')





import win32clipboard
from cStringIO import StringIO
from PIL import ImageGrab as PilImageGrab


class WindowsBackend(BaseBackend):
    """ Windows clipboard backend

    This backends the windows clipboard via the win32clipboard module.
    This means that there is less code to write, because a fair amount
    of it is already written. This backend really just takes its functionality
    and makes it compatible with the frontend. Also, it uses the PIL.ImageGrab
    module to get images from the clipboard, so PIL.Image is considered the
    native image type
    """

    @staticmethod
    def get_native_image_str():
        return 'pil'

    @staticmethod
    def get_native_image_type():
        return PilImage

    def __init__(self):
        """ Constructs a new object

        There isn't really an internal clipboard object to manage. Instead, just
        open the clipboard
        """
        win32clipboard.OpenClipboard()

    def __del__(self):
        win32clipboard.CloseClipboard()

    def get_text(self):   
        """ Synchronously get text

        Preform a blocking call to get text from the clipboard

        Returns
        -------
        str or None
            String if text is ever gotten, or None if there is no text
        """
        return win32clipboard.GetClipboardData()

    def get_image(self, format='Pil'):
        """ Synchronously get image

        Preform a blocking call to get an image from the clipboard

        Params
        ----------
        format : str (optional)
            Format of the image. Default is 'Pil', which refers to
            PIL.Image, but implementation is left to each function

        Returns
        -------
        PIL.Image or None
            Image from clipboard, or None if not available
        """
        return PilImageGrab.grabclipboard()

    def set_text(self, text):
        """ Synchronously sets text

        Preform a blocking call to set text on the clipboard

        Params
        ----------
        text : str
            Text to add to clipboard
        """
        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboard(text.encode('utf-8'), win32clipboard.CF_TEXT)

    def set_image(self, img, format='pil'):
        """ Synchronously sets image

        Preform a blocking call to set image on the clipboard

        Params
        ----------
        image : PIL.Image
            Image to add to clipboard
        """
        output = StringIO()
        img.convert("RGB").save(output, 'BMP')
        data = output.getvalue()[14:]
        output.close()

        win32clipboard.EmptyClipboard()
        win32clipboard.SetClipboardData(win32clipboard.CF_DIB, data)




# crossclip -- cross platform clipboard API
# Copyright (C) 2019  Charlie Sale

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# qtbackend.py -- qt backend class

from PyQt5.Qt import QApplication, QClipboard, QBuffer
from PyQt5.QtGui import QImage, QPixmap
import PyQt5
from PIL import Image as PilImage

from .absbackend import AbstractBackend, AbstractImageConverter


class QtImageConverter(AbstractImageConverter):

    native_image = QImage
    native_image_str = 'qt'

    def to_pillow(qimage):
        buf = QBuffer()
        buf.open(QBuffer.ReadWrite)
        qimage.save(buf, 'PNG') # TODO: make this dynamic
        pimage = PilImage.open(io.BytesIO(buf.data()))
        return pimage

    def from_pillow(image):
        qimg = PilImageQt(image)
        return qimg

class QtBackend(AbstractBackend):
    """ Backend for Qt clipboard

    This class backends the default Qt Clipboard
    """
    image_converter = QtImageConverter

    def __init__(self):
        # Get the default application. I am ignoring any sort
        # of signal/slot setup. This will all be based off of user
        # actions
        super().__init__()
        self.app = QApplication([])
        self.clipboard = self.app.clipboard()

    def get_text(self):
        return self.clipboard.text()

    def get_image(self, format='pil'):
        img = self.clipboard.image()
        if format == self.image_converter.native_image_str:
            return img
        elif format == 'pil':
            return self.image_converter.to_pillow(img)
        else:
            raise RuntimeWarning('Image format is not supported')
            return None

    def set_text(self, text):
        self.clipboard.setText(text)

    def set_image(self, image, format='pil'):
        if format == self.image_converter.native_image_str:
            self.clipboard.setImage(image)
        elif format == 'pil':
            qimg = self.image_converter.from_pillow(image)
            self.clipboard.setImage(qimg)
        else:
            raise RuntimeWarning('Image format is not supported')
            return None



#! /usr/bin/env python3

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="crossclip",
    version="0.2.2",
    author="Charlie Sale",
    author_email="softwaresale01@gmail.com",
    description="A Cross Platform clipboard manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/softwaresale/crossclip",
    packages=setuptools.find_packages(),
    install_requires=[
        'Pillow',
    ],
    extras_require={
        'GTK': ['pygobject'],
        'Qt': ['PyQt5'],
    },
    test_suite='nose.collector',
    tests_require=['nose', 'numpy'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
)




# crossclip -- cross platform clipboard API
# Copyright (C) 2019  Charlie Sale

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# gtkbackend.py -- Defines GTK backend

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GdkPixbuf
    from gi.repository import GLib
except ModuleNotFoundError:
    print("pygobject cannot be found. Please install the pip package: 'pygobject'")

from PIL import Image as PilImage
from PIL.Image import Image as PilImageType
from io import BytesIO

from .absbackend import AbstractBackend, AbstractImageConverter

class GtkImageConverter(AbstractImageConverter):

    @property
    def image_type(self):
        """
        Returns the type of image that this converter uses.

        :returns: GdkPixbuf.Pixbuf type (not object!)
        """
        return GdkPixbuf.Pixbuf

    @property
    def image_str(self):
        """
        Returns a string representation of what the object is.

        :returns str: 'gdk-pixbuf'
        """
        return 'gdk-pixbuf'

    def to_pillow(self, pixbuf):
        """
        Converts and image of `self.image_type` to a `PIL.Image`.

        :param pixbuf: GdkPixbuf.Pixbuf image
        :returns PIL.Image: Converted Pillow Image
        """
        # Sanity check to make sure that pixbuf isn't already a Pillow Image
        if isinstance(pixbuf, PilImageType):
            return pixbuf

        data = pixbuf.get_pixels()
        w = pixbuf.props.width
        h = pixbuf.props.height
        stride = pixbuf.props.rowstride
        mode = "RGB"
        if pixbuf.props.has_alpha == True:
            mode = "RGBA"
        return PilImage.frombytes(mode, (w, h), data, "raw", mode, stride)

    def from_pillow(self, image):
        """
        Converts a `PIL.Image` to a `GdkPixbuf.Pixbuf`.

        :param image: `PIL.Image` to be convered
        :returns GdkPixbuf.Pixbuf: Converted pixbuf
        """

        # data = image.tobytes()
        # w, h = image.size
        # data = GLib.Bytes.new(data)
        # pix = GdkPixbuf.Pixbuf.new_from_bytes(data, GdkPixbuf.Colorspace.RGB, False, 8, w, h, w * 3)
        # return pix
        # Sanity check to verify that image isn't already native type
        if isinstance(image, self.image_type):
            return image

        ibuf = BytesIO()
        image.save(ibuf, format='png')
        loader = GdkPixbuf.PixbufLoader.new_with_mime_type('image/png')
        status = loader.write(ibuf.getvalue())
        if status:
            pixbuf = loader.get_pixbuf()
            loader.close()
            return pixbuf
        else:
            return None

class GtkBackend(AbstractBackend):
    """ Gtk Clipboard backend

    This clipboard backends the Gtk.Clipboard class. It inherits from it and
    the base backend class. This way, it has all of the Gtk's properties, but
    it can operate through the BaseBackend interface.
    """

    image_converter = GtkImageConverter()
    raw_clipboard = None

    def __init__(self, display=None):
        if display is None:
            display = Gdk.Display.get_default()
        super().__init__()
        self.clipboard = Gtk.Clipboard.get_default(display)
        self.raw_clipboard = self.clipboard

    def get_text(self):
        """
        Synchronously gets text from clipboard.

        :return str: Text from clipboard
        """
        text = self.clipboard.wait_for_text()
        return text

    def get_image(self, format='pil', converter=None):
        """
        Synchronously gets image from clipboard. The image is either
        a pillow image, or a GdkPixbuf.Pixbuf.

        :param format: Format of image. 'pil' for pillow, 'gdk-pixbuf' for gdk pixbuf (default: 'pil')
        :returns PIL.Image or GdkPixbuf.Pixbuf: Image in chosen format
        :raises RuntimeWarning: If format is invalid format
        """
        pixbuf = self.clipboard.wait_for_image()
        if pixbuf is None:
            return None

        if format == self.image_converter.image_str:
            return pixbuf
        elif format == 'pil':
            return self.image_converter.to_pillow(pixbuf)
        else:
            if converter is not None and isinstance(converter, AbstractImageConverter):
                pil = self.image_converter.to_pillow(pixbuf)
                return converter.from_pillow(pil)
            else:
                raise RuntimeWarning("Invalid format, and converter is not provided")

    def set_text(self, text, num=-1):
        """
        Synchronously sets text to clipboard

        :param text: text to set to clipboard
        :param num: length of text to set. If -1, all text is copied (default: -1)
        """
        # Assuming that all text is to be copied over
        self.clipboard.set_text(text, num)
        self.clipboard.store()

    def set_image(self, image, converter=None):
        """
        Synchronously sets image to clipboard.

        :param image: Some image object
        :param format: format of image
        :raises RuntimeWarning: If format is invalid
        """
        # if format == self.image_converter.image_str:
        #     self.clipboard.set_image(image)
        #     self.clipboard.store()
        # elif format == 'pil':
        #     pixbuf = self.image_converter.from_pillow(image)
        #     self.clipboard.set_image(pixbuf)
        #     self.clipboard.store()
        # else:
        #     raise RuntimeWarning('Invalid image format')
        #     return None
        if isinstance(image, PilImageType):
            # If image is a pillow image, then it needs to be converted
            pixbuf = self.image_converter.from_pillow(image)
            self.clipboard.set_image(pixbuf)
            self.clipboard.store()
        elif isinstance(image, self.image_converter.image_type):
            # Image is already native type, good to go
            self.clipboard.set_image(pixbuf)
            self.clipboard.store()
        else:
            # If a converter is provided, then use it to convert the image to a
            # Pillow object and recursively run the method.
            if converter is not None and isinstance(converter, AbstractImageConverter):
                pillow_img = converter.to_pillow(image)
                self.set_image(pillow_img)
            else:
                raise RuntimeWarning("Image is of invalid type and has no converter")
